from __future__ import absolute_import
from ._version import __version__

from . import errors
from . import parsers
from . import response
from . import resource
from . import routing
from . import schema

__all__ = [__version__, errors, parsers, response, resource, routing, schema]



import inspect
import os
from functools import wraps
from inspect import Parameter

import mock

from doctor.routing import get_params_from_func
from doctor.utils import (
    add_param_annotations, get_description_lines, get_module_attr,
    get_valid_class_name, Params, RequestParamAnnotation)

from .base import TestCase
from .types import Age, Auth, Foo, IsAlive, IsDeleted, Name


def does_nothing(func):
    """An example decorator that does nothing."""
    @wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)
    return wrapper


def dec_with_args(foo='foo', bar='bar'):
    """An example decorator that takes args and passes value to func."""
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            return func('arg', *args, **kwargs)
        return wrapper
    return decorator


def dec_with_args_one_of_which_allows_a_func(foo, bar=None):
    """An example decorator that takes args where one is a function.

    In this case, bar accepts a func which transforms foo.
    """
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if bar is not None:
                bar(foo)
            return func(*args, **kwargs)
        return wrapper
    return decorator


def no_params() -> Foo:
    return ''


def get_foo(name: Name, age: Age, is_alive: IsAlive=True) -> Foo:
    return ''


def pass_pos_param(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        return func('extra!', *args, **kwargs)
    return wrapper


@pass_pos_param
def decorated_func(extra: str, name: Name, is_alive: IsAlive=True) -> Foo:
    return ''


class TestUtils(TestCase):

    def test_add_param_annotations(self):
        new_params = [
            RequestParamAnnotation('auth', Auth, required=True),
            RequestParamAnnotation('is_deleted', IsDeleted)
        ]
        actual = add_param_annotations(get_foo, new_params)
        # auth and is_deleted should be added to `all`.
        # auth should be added to `required`.
        # is_deleted should be added to `optional`.
        # `logic` should be unmodified.
        expected_params = Params(
            all=['name', 'age', 'is_alive', 'auth', 'is_deleted'],
            logic=['name', 'age', 'is_alive'],
            required=['name', 'age', 'auth'],
            optional=['is_alive', 'is_deleted']
        )
        assert expected_params == actual._doctor_params

        # verify `auth` added to the doctor_signature.
        expected = Parameter('auth', Parameter.KEYWORD_ONLY,
                             default=Parameter.empty, annotation=Auth)
        auth = actual._doctor_signature.parameters['auth']
        assert expected == auth

        # verify `is_deleted` added to the doctor signature.
        expected = Parameter('is_deleted', Parameter.KEYWORD_ONLY,
                             default=None, annotation=IsDeleted)
        is_deleted = actual._doctor_signature.parameters['is_deleted']
        assert expected == is_deleted

    def test_add_param_annotations_mutliple_calls(self):
        """
        This test verifies if we call this function multiple times with
        the same logic function that it doesn't squash parameters added
        from the previous call.  This is a regression test.
        """
        delattr(get_foo, '_doctor_signature')
        new_params = [
            RequestParamAnnotation('auth', Auth, required=True),
        ]
        actual = add_param_annotations(get_foo, new_params)
        expected_params = Params(
            all=['name', 'age', 'is_alive', 'auth'],
            logic=['name', 'age', 'is_alive'],
            required=['name', 'age', 'auth'],
            optional=['is_alive']
        )
        assert expected_params == actual._doctor_params

        # verify `auth` added to the doctor_signature.
        expected = Parameter('auth', Parameter.KEYWORD_ONLY,
                             default=Parameter.empty, annotation=Auth)
        auth = actual._doctor_signature.parameters['auth']
        assert expected == auth

        # Now call again adding the is_deleted param.  It should add to and
        # not replace the auth param we added previously.
        new_params = [
            RequestParamAnnotation('is_deleted', IsDeleted)
        ]
        actual = add_param_annotations(get_foo, new_params)
        expected_params = Params(
            all=['name', 'age', 'is_alive', 'auth', 'is_deleted'],
            logic=['name', 'age', 'is_alive'],
            required=['name', 'age', 'auth'],
            optional=['is_alive', 'is_deleted']
        )
        assert expected_params == actual._doctor_params
        # verify `is_deleted` added to the doctor signature.
        expected = Parameter('is_deleted', Parameter.KEYWORD_ONLY,
                             default=None, annotation=IsDeleted)
        is_deleted = actual._doctor_signature.parameters['is_deleted']
        assert expected == is_deleted

    @mock.patch('doctor.utils.logging')
    def test_add_param_annotations_duplicate_param(self, mock_logging):
        delattr(get_foo, '_doctor_signature')
        new_params = [
            # name param is already in `get_foo` signature
            RequestParamAnnotation('name', Name, required=True),
            RequestParamAnnotation('is_deleted', IsDeleted)
        ]
        actual = add_param_annotations(get_foo, new_params)

        # verify `is_deleted` added to the doctor signature.
        expected = Parameter('is_deleted', Parameter.KEYWORD_ONLY,
                             default=None, annotation=IsDeleted)
        is_deleted = actual._doctor_signature.parameters['is_deleted']
        assert expected == is_deleted

        # verify a warning was logged.
        expected_call = mock.call(
            'Not adding %s to signature of %s, function already has that '
            'parameter in its signature.', 'name', 'get_foo')
        assert expected_call == mock_logging.warning.call_args

    @mock.patch('doctor.utils.open',
                new_callable=mock.mock_open,
                read_data='mock_attr = "something"')
    @mock.patch('doctor.utils.compile', create=True)
    @mock.patch('doctor.utils.os', autospec=True)
    @mock.patch('doctor.utils.sys', autospec=True)
    def test_get_module_attr(self, mock_sys, mock_os, mock_compile, m):
        def side_effect(source, filename, mode, flags=0, dont_inherit=False):
            assert mock_sys.path == ['foo', 'bar', '/foo/bar']
            return compile(source, filename, mode, flags=flags,
                           dont_inherit=dont_inherit)
        mock_compile.side_effect = side_effect
        mock_os.getcwd.return_value = mock.sentinel.old_cwd
        mock_os.path = os.path
        mock_sys.path = ['foo', 'bar']
        namespace = {}
        result = get_module_attr('/foo/bar/baz', 'mock_attr',
                                 namespace=namespace)
        m.assert_called_once_with('/foo/bar/baz', 'r')
        mock_compile.assert_called_once_with('mock_attr = "something"',
                                             '/foo/bar/baz', 'exec')
        assert result == 'something'
        assert namespace['__file__'] == '/foo/bar/baz'
        assert mock_os.chdir.call_args_list == [
            mock.call('/foo/bar'),
            mock.call(mock.sentinel.old_cwd),
        ]
        assert mock_sys.path == ['foo', 'bar']

    def test_get_description_lines(self):
        """
        Tests that get_description_lines properly dedents docstrings and strips
        out values we don't want.
        """
        docstring = """line one

                       line two

                           indented line three

                       line four

                       :param str c: example param
                       :returns: d
                       """
        assert get_description_lines(docstring) == [
            'line one',
            '',
            'line two',
            '',
            '    indented line three',
            '',
            'line four',
            '',
        ]

    def test_get_description_lines_none(self):
        """It should just return an empty list for None."""
        assert get_description_lines(None) == []

    def test_get_description_lines_trailing_newline(self):
        """It should add a trailing line if necessary."""
        assert get_description_lines('foo\n:arg') == ['foo', '']

    def test_get_params_from_func(self):
        get_foo._doctor_signature = inspect.signature(get_foo)
        expected = Params(
            all=['name', 'age', 'is_alive'],
            optional=['is_alive'],
            required=['name', 'age'],
            logic=['name', 'age', 'is_alive'])
        assert expected == get_params_from_func(get_foo)

    def test_get_params_from_func_no_params(self):
        # no signature passed or defined on the function
        expected = Params([], [], [], [])
        assert expected == get_params_from_func(no_params)

        # signature passed in
        signature = inspect.signature(no_params)
        assert expected == get_params_from_func(no_params, signature)

        # signature attached to logic function
        no_params._doctor_signature = inspect.signature(no_params)
        assert expected == get_params_from_func(no_params)

    def test_get_params_from_func_decorated_func(self):
        """
        Verifies that we don't include the `extra` param as required since
        it's not a sublcass of `SuperType` and is passed to the function
        by a decorator.
        """
        decorated_func._doctor_signature = inspect.signature(decorated_func)
        expected = Params(
            all=['extra', 'name', 'is_alive'],
            required=['name'],
            optional=['is_alive'],
            logic=['extra', 'name', 'is_alive'])
        assert expected == get_params_from_func(decorated_func)

    def test_get_valid_class_name(self):
        tests = (
            # (input, expected)
            ('Notes', 'Notes'),
            ('Notes (v1)', 'NotesV1'),
            ('Notes - "v1"', 'NotesV1'),
            ('note_book.', 'NoteBook'),
            ('notes', 'Notes'),
            ('note-book_v1 .', 'NoteBookV1'),
        )
        for s, expected in tests:
            assert expected == get_valid_class_name(s)



from typing import Callable

from .schema import Schema
from .types import SuperType


#: A mapping of HTTP method to title that should be used for it in
#: API documentation.
HTTP_METHOD_TITLES = {
    'DELETE': 'Delete',
    'GET': 'Retrieve',
    'POST': 'Create',
    'PUT': 'Update',
}


class ResourceSchemaAnnotation(object):

    """Metadata about the schema used for a given request method.

    An instance of this class is attached to each handler method in a
    _schema_annotation attribute. It can be used for introspection about the
    schemas, to generate things like API documentation and hyper schemas from
    the code.

    :param func logic: Logic function which will handle the request.
    :param str http_method: The HTTP request method for this request (e.g. GET).
    :param doctor.resource.ResourceSchema schema: The resource schema
        object for this handler.
    :param dict request_schema: The schema used to validate the request.
    :param dict response_schema: The schema used to validate the response.
    :param str title: A short title for the route.  e.g. 'Create Foo' might
        be used for a POST method on a FooListHandler.
    """

    def __init__(self, logic, http_method, schema, request_schema,
                 response_schema, title=None):
        self.logic = logic
        self.http_method = http_method
        self.schema = schema
        self.request_schema = request_schema
        self.response_schema = response_schema
        self.title = title
        if title is None:
            self.title = HTTP_METHOD_TITLES.get(http_method.upper())

    @classmethod
    def get_annotation(cls, fn):
        """Find the _schema_annotation attribute for the given function.

        This will descend through decorators until it finds something that has
        the attribute. If it doesn't find it anywhere, it will return None.

        :param func fn: Find the attribute on this function.
        :returns: an instance of
            :class:`~doctor.resource.ResourceSchemaAnnotation` or
            None.
        """
        while fn is not None:
            if hasattr(fn, '_schema_annotation'):
                return fn._schema_annotation
            fn = getattr(fn, 'im_func', fn)
            closure = getattr(fn, '__closure__', None)
            fn = closure[0].cell_contents if closure is not None else None
        return None


class ResourceAnnotation(object):

    """Metadata about the types used for a given request method.

    :param logic: The logic function for the resource.
    :param http_method: The http method for this resource.  One of `DELETE`,
        `GET`, `POST` or `PUT`.
    :param title: The title to be used by the api documentation for this
        resource.
    """

    def __init__(self, logic: Callable, http_method: str, title: str=None):
        self.annotated_parameters = {
            k: p for k, p in logic._doctor_signature.parameters.items()
            if issubclass(p.annotation, SuperType)
        }
        self.http_method = http_method.upper()
        self.logic = logic
        self.params = logic._doctor_params
        self.return_annotation = logic._doctor_signature.return_annotation
        self.title = title
        if title is None:
            self.title = HTTP_METHOD_TITLES.get(http_method.upper())


class ResourceSchema(Schema):

    """
    This class extends :class:`~doctor.schema.Schema` with methods
    for generating HTTP handler functions that automatically parse and
    validate the request and response objects with a given schema.
    """

    def __init__(self, schema, handle_http=None,
                 raise_response_validation_errors=False, **kwargs):
        """
        :param dict schema: The JSON schema to use for this resource.
        :param function handle_http: The HTTP handler function that should be
            used to wrap the logic functions. You would normally pass
            :func:`doctor.flask.handle_http`.
        :param bool raise_response_validation_errors: True to raise errors
            for response validation exceptions, False to just log them
            and return gracefully.
        """
        super(ResourceSchema, self).__init__(schema, **kwargs)
        self.handle_http = handle_http
        self.raise_response_validation_errors = raise_response_validation_errors

    def _create_request_schema(self, params, required):
        """Create a JSON schema for a request.

        :param list params: A list of keys specifying which definitions from
            the base schema should be allowed in the request.
        :param list required: A subset of the params that the requester must
            specify in the request.
        :returns: a JSON schema dict
        """
        # We allow additional properties because the data this will validate
        # may also include kwargs passed by decorators on the handler method.
        schema = {'additionalProperties': True,
                  'definitions': self.resolve('#/definitions'),
                  'properties': {},
                  'required': required or (),
                  'type': 'object'}
        for param in params:
            schema['properties'][param] = {
                '$ref': '#/definitions/{}'.format(param)}
        return schema



from __future__ import absolute_import

import logging
import os
from typing import Callable, Dict, List, Tuple, Union


try:
    from flask import current_app, request
    from flask_restful import Resource
    from werkzeug.exceptions import (BadRequest, Conflict, Forbidden,
                                     HTTPException, NotFound, Unauthorized,
                                     InternalServerError)
except ImportError:  # pragma: no cover
    raise ImportError('You must install flask to use the '
                      'doctor.flask module.')

from .constants import HTTP_METHODS_WITH_JSON_BODY
from .errors import (ForbiddenError, ImmutableError, InvalidValueError,
                     NotFoundError, TypeSystemError, UnauthorizedError)
from .parsers import map_param_names, parse_form_and_query_params
from .response import Response
from .routing import create_routes as doctor_create_routes
from .routing import Route


STATUS_CODE_MAP = {
    'POST': 201,
    'DELETE': 204,
}

ListOrNone = Union[List, None]


class SchematicHTTPException(HTTPException):

    """Schematic specific sub-class of werkzeug's BadRequest.

    Note that this adds a flask-restful specific data attribute to the class,
    as the error wouldn't render properly without it.

    :param description: The error description.
    :param errors: A dict containing all validation errors during the request.
        The key is the param name and the value is the error message.
    """

    def __init__(self, description: str=None, errors: dict=None):
        super(SchematicHTTPException, self).__init__(description)
        self.data = {'status': self.code, 'message': description}
        self.errors = errors

    def __str__(self):
        return '%d: %s: %s' % (self.code, self.name, self.description)


class HTTP400Exception(SchematicHTTPException, BadRequest):
    """Represents a HTTP 400 error.

    :param description: The error description.
    :param errors: A dict containing all validation errors during the request.
        The key is the param name and the value is the error message.
    """
    pass


class HTTP401Exception(SchematicHTTPException, Unauthorized):
    pass


class HTTP403Exception(SchematicHTTPException, Forbidden):
    pass


class HTTP404Exception(SchematicHTTPException, NotFound):
    pass


class HTTP409Exception(SchematicHTTPException, Conflict):
    pass


class HTTP500Exception(SchematicHTTPException, InternalServerError):
    pass


def should_raise_response_validation_errors() -> bool:
    """Returns if the library should raise response validation errors or not.

    If the environment variable `RAISE_RESPONSE_VALIDATION_ERRORS` is set,
    it will return True.

    :returns: True if it should, False otherwise.
    """
    return bool(os.environ.get('RAISE_RESPONSE_VALIDATION_ERRORS', False))


def handle_http(handler: Resource, args: Tuple, kwargs: Dict, logic: Callable):
    """Handle a Flask HTTP request

    :param handler: flask_restful.Resource: An instance of a Flask Restful
        resource class.
    :param tuple args: Any positional arguments passed to the wrapper method.
    :param dict kwargs: Any keyword arguments passed to the wrapper method.
    :param callable logic: The callable to invoke to actually perform the
        business logic for this request.
    """
    try:
        # We are checking mimetype here instead of content_type because
        # mimetype is just the content-type, where as content_type can
        # contain encoding, charset, and language information.  e.g.
        # `Content-Type: application/json; charset=UTF8`
        if (request.mimetype == 'application/json' and
                request.method in HTTP_METHODS_WITH_JSON_BODY):
            # This is a proper typed JSON request. The parameters will be
            # encoded into the request body as a JSON blob.
            if not logic._doctor_req_obj_type:
                request_params = map_param_names(
                    request.json, logic._doctor_signature.parameters)
            else:
                request_params = request.json
        else:
            # Try to parse things from normal HTTP parameters
            request_params = parse_form_and_query_params(
                request.values, logic._doctor_signature.parameters)

        params = request_params
        # Only filter out additional params if a req_obj_type was not specified.
        if not logic._doctor_req_obj_type:
            # Filter out any params not part of the logic signature.
            all_params = logic._doctor_params.all
            params = {k: v for k, v in params.items() if k in all_params}
        params.update(**kwargs)

        # Check for required params
        missing = []
        for required in logic._doctor_params.required:
            if required not in params:
                missing.append(required)
        if missing:
            verb = 'are'
            if len(missing) == 1:
                verb = 'is'
                missing = missing[0]
            error = '{} {} required.'.format(missing, verb)
            raise InvalidValueError(error)

        # Validate and coerce parameters to the appropriate types.
        errors = {}
        sig = logic._doctor_signature
        # If a `req_obj_type` was defined for the route, pass all request
        # params to that type for validation/coercion
        if logic._doctor_req_obj_type:
            annotation = logic._doctor_req_obj_type
            try:
                # NOTE: We calculate the value before applying native type in
                # order to support UnionType types which dynamically modifies
                # the native_type property based on the initialized value.
                value = annotation(params)
                params = annotation.native_type(value)
            except TypeError:
                logging.exception(
                    'Error casting and validating params with value `%s`.',
                    params)
                raise
            except TypeSystemError as e:
                errors['__all__'] = e.detail
        else:
            for name, value in params.items():
                annotation = sig.parameters[name].annotation
                if annotation.nullable and value is None:
                    continue
                try:
                    # NOTE: We calculate the value before applying native type
                    # in order to support UnionType types which dynamically
                    # modifies the native_type property based on the initialized
                    # value.
                    value = annotation(value)
                    params[name] = annotation.native_type(value)
                except TypeSystemError as e:
                    errors[name] = e.detail
        if errors:
            raise TypeSystemError(errors, errors=errors)

        if logic._doctor_req_obj_type:
            # Pass any positional arguments followed by the coerced request
            # parameters to the logic function.
            response = logic(*args, params)
        else:
            # Only pass request parameters defined by the logic signature.
            logic_params = {k: v for k, v in params.items()
                            if k in logic._doctor_params.logic}
            response = logic(*args, **logic_params)

        # response validation
        if sig.return_annotation != sig.empty:
            return_annotation = sig.return_annotation
            _response = response
            if isinstance(response, Response):
                _response = response.content
                # Check if our return annotation is a Response that supplied a
                # type to validate against.  If so, use that type for validation
                # e.g. def logic() -> Response[MyType]
                if (issubclass(return_annotation, Response) and
                        return_annotation.__args__ is not None):
                    return_annotation = return_annotation.__args__[0]
            try:
                return_annotation(_response)
            except TypeSystemError as e:
                response_str = str(_response)
                logging.warning('Response to %s %s does not validate: %s.',
                                request.method, request.path,
                                response_str, exc_info=e)
                if should_raise_response_validation_errors():
                    error = ('Response to {method} {path} `{response}` does not'
                             ' validate: {error}'.format(
                                 method=request.method, path=request.path,
                                 response=response, error=e.detail))
                    raise TypeSystemError(error)

        if isinstance(response, Response):
            status_code = response.status_code
            if status_code is None:
                status_code = STATUS_CODE_MAP.get(request.method, 200)
            return (response.content, status_code, response.headers)
        return response, STATUS_CODE_MAP.get(request.method, 200)
    except (InvalidValueError, TypeSystemError) as e:
        errors = getattr(e, 'errors', None)
        raise HTTP400Exception(e, errors=errors)
    except UnauthorizedError as e:
        raise HTTP401Exception(e)
    except ForbiddenError as e:
        raise HTTP403Exception(e)
    except NotFoundError as e:
        raise HTTP404Exception(e)
    except ImmutableError as e:
        raise HTTP409Exception(e)
    except Exception as e:
        # Always re-raise exceptions when DEBUG is enabled for development.
        if current_app.config.get('DEBUG', False):
            raise
        allowed_exceptions = logic._doctor_allowed_exceptions
        if allowed_exceptions and any(isinstance(e, cls)
                                      for cls in allowed_exceptions):
            raise
        logging.exception(e)
        raise HTTP500Exception('Uncaught error in logic function')


def create_routes(routes: Tuple[Route]) -> List[Tuple[str, Resource]]:
    """A thin wrapper around create_routes that passes in flask specific values.

    :param routes: A tuple containing the route and another tuple with
        all http methods allowed for the route.
    :returns: A list of tuples containing the route and generated handler.
    """
    return doctor_create_routes(
        routes, handle_http, default_base_handler_class=Resource)



import logging
import os

import jsonschema
import yaml
from jsonschema.compat import urldefrag

from .errors import (
    DoctorError, SchemaError, SchemaLoadingError, SchemaValidationError)
from .parsers import parse_json


DEFAULT = object()


class SchemaRefResolver(jsonschema.RefResolver):

    """Subclass in order to provide support for loading YAML files."""

    def _format_stack(self, stack, current=None):
        """Prettifies a scope stack for use in error messages.

        :param list(str) stack: List of scopes.
        :param str current: The current scope. If specified, will be appended
            onto the stack before formatting.
        :returns: str
        """
        if current is not None:
            stack = stack + [current]
        if len(stack) > 1:
            prefix = os.path.commonprefix(stack)
            if prefix.endswith('/'):
                prefix = prefix[:-1]
            stack = [scope[len(prefix):] for scope in stack]
        return ' => '.join(stack)

    def resolve(self, ref, document=None):
        """Resolve a fragment within the schema.

        If the resolved value contains a $ref, it will attempt to resolve that
        as well, until it gets something that is not a reference. Circular
        references will raise a SchemaError.

        :param str ref: URI to resolve.
        :param dict document: Optional schema in which to resolve the URI.
        :returns: a tuple of the final, resolved URI (after any recursion) and
            resolved value in the schema that the URI references.
        :raises SchemaError:
        """
        try:
            # This logic is basically the RefResolver's resolve function, but
            # updated to support fragments of dynamic documents. The jsonschema
            # module supports passing documents when resolving fragments, but
            # it doesn't expose that capability in the resolve function.
            url = self._urljoin_cache(self.resolution_scope, ref)
            if document is None:
                # No document passed, so just resolve it as we normally would.
                resolved = self._remote_cache(url)
            else:
                # Document passed, so assume it's a fragment.
                _, fragment = urldefrag(url)
                resolved = self.resolve_fragment(document, fragment)
        except jsonschema.RefResolutionError as e:
            # Failed to find a ref. Make the error a bit prettier so we can
            # figure out where it came from.
            message = e.args[0]
            if self._scopes_stack:
                message = '{} (from {})'.format(
                    message, self._format_stack(self._scopes_stack))
            raise SchemaError(message)

        if isinstance(resolved, dict) and '$ref' in resolved:
            # Try to resolve the reference, so we can get the actual value we
            # want, instead of a useless dict with a $ref in it.
            if url in self._scopes_stack:
                # We've already tried to look up this URL, so this must
                # be a circular reference in the schema.
                raise SchemaError(
                    'Circular reference in schema: {}'.format(
                        self._format_stack(self._scopes_stack + [url])))
            try:
                self.push_scope(url)
                return self.resolve(resolved['$ref'])
            finally:
                self.pop_scope()
        else:
            return url, resolved

    def resolve_remote(self, uri):
        """Add support to load YAML files.

        This will attempt to load a YAML file first, and then go back to the
        default behavior.

        :param str uri: the URI to resolve
        :returns: the retrieved document
        """
        if uri.startswith('file://'):
            try:
                path = uri[7:]
                with open(path, 'r') as schema_file:
                    result = yaml.load(schema_file)
                if self.cache_remote:
                    self.store[uri] = result
                return result
            except yaml.parser.ParserError as e:
                logging.debug('Error parsing {!r} as YAML: {}'.format(
                    uri, e))
        return super(SchemaRefResolver, self).resolve_remote(uri)


class Schema(object):

    """
    This class is used to manipulate JSON schemas and validate values against
    the schema.

    :param dict schema: The loaded schema.
    :param str schema_path: The absolute path to the directory of local schemas.
    """

    def __init__(self, schema, schema_path=None):
        self.schema = schema
        self._resolver = None
        self._schema_path = schema_path

    def get_validator(self, schema=None):
        """Get a jsonschema validator.

        :param dict schema: A custom schema to validate against.
        :returns: an instance of jsonschema Draft4Validator.
        """
        schema = schema if schema is not None else self.schema
        return jsonschema.Draft4Validator(
            schema, resolver=self.resolver,
            format_checker=jsonschema.draft4_format_checker)

    def resolve(self, ref, document=None):
        """Resolve a ref within the schema.

        This is just a convenience method, since RefResolver returns both a URI
        and the resolved value, and we usually just need the resolved value.

        :param str ref: URI to resolve.
        :param dict document: Optional schema in which to resolve the URI.
        :returns: the portion of the schema that the URI references.
        :see: :meth:`SchemaRefResolver.resolve`
        """
        _, resolved = self.resolver.resolve(ref, document=document)
        return resolved

    @property
    def resolver(self):
        """jsonschema RefResolver object for the base schema."""
        if self._resolver is not None:
            return self._resolver
        if self._schema_path is not None:
            # the documentation for ref resolving
            # https://github.com/Julian/jsonschema/issues/98
            # https://python-jsonschema.readthedocs.org/en/latest/references/
            self._resolver = SchemaRefResolver(
                'file://' + self._schema_path + '/', self.schema)
        else:
            self._resolver = SchemaRefResolver.from_schema(self.schema)
        return self._resolver

    def validate(self, value, validator):
        """Validates and returns the value.

        If the value does not validate against the schema, SchemaValidationError
        will be raised.

        :param value: A value to validate (usually a dict).
        :param validator: An instance of a jsonschema validator class, as
            created by Schema.get_validator().
        :returns: the passed value.
        :raises SchemaValidationError:
        :raises Exception:
        """
        try:
            validator.validate(value)
        except Exception as e:
            logging.debug(e, exc_info=e)
            if isinstance(e, DoctorError):
                raise
            else:
                # Gather all the validation errors
                validation_errors = sorted(
                    validator.iter_errors(value), key=lambda e: e.path)
                errors = {}
                for error in validation_errors:
                    try:
                        key = error.path[0]
                    except IndexError:
                        key = '_other'
                    errors[key] = error.args[0]
                raise SchemaValidationError(e.args[0], errors=errors)
        return value

    def validate_json(self, json_value, validator):
        """Validates and returns the parsed JSON string.

        If the value is not valid JSON, ParseError will be raised. If it is
        valid JSON, but does not validate against the schema,
        SchemaValidationError will be raised.

        :param str json_value: JSON value.
        :param validator: An instance of a jsonschema validator class, as
            created by Schema.get_validator().
        :returns: the parsed JSON value.
        """
        value = parse_json(json_value)
        return self.validate(value, validator)

    @classmethod
    def from_file(cls, schema_filepath, *args, **kwargs):
        """Create an instance from a YAML or JSON schema file.

        Any additional args or kwargs will be passed on when constructing the
        new schema instance (useful for subclasses).

        :param str schema_filepath: Path to the schema file.
        :returns: an instance of the class.
        :raises SchemaLoadingError: for invalid input files.
        """
        schema_filepath = os.path.abspath(schema_filepath)
        try:
            with open(schema_filepath, 'r') as schema_file:
                schema = yaml.load(schema_file.read())
        except Exception:
            msg = 'Error loading schema file {}'.format(schema_filepath)
            logging.exception(msg)
            raise SchemaLoadingError(msg)
        return cls(schema, *args, schema_path=os.path.dirname(schema_filepath),
                   **kwargs)



from typing import Generic, TypeVar


#: A type variable to represent the type of content of a `Response`.
CT = TypeVar('CT')


class Response(Generic[CT]):
    """Represents a response.

    This object contains the response itself along with any additional headers
    that should be added and returned with the response data.  An instance of
    this class can be returned from a logic function in order to modify
    response headers.

    :param content: The data to be returned with the response.
    :param dict headers: A dict of response headers to include with the response
    :param int status_code: The status code for the response.
    """

    def __init__(self, content: CT, headers: dict = None,
                 status_code: int = None):
        self.content = content
        self.headers = headers
        self.status_code = status_code



#: HTTP methods that are allowed to have a JSON body.  Technically all HTTP
#: methods are allowed to have a body, but some like GET/DELETE have no
#: contextual meaning server side, so should not be used.
HTTP_METHODS_WITH_JSON_BODY = ('PATCH', 'POST', 'PUT')



import functools
import inspect
from typing import Any, Callable, List, Tuple

from doctor.utils import copy_func, get_params_from_func, get_valid_class_name


class HTTPMethod(object):
    """Represents and HTTP method and it's configuration.

    When instantiated the logic attribute will have 3 attributes added to it:
        - `_doctor_allowed_exceptions` - A list of excpetions that are allowed
          to be re-reaised if encountered during a request.
        - `_doctor_params` - A :class:`~doctor.utils.Params` instance.
        - `_doctor_signature` - The parsed function Signature.
        - `_doctor_title` - The title that should be used in api documentation.

    :param method: The HTTP method.  One of: (delete, get, post, put).
    :param logic: The logic function to be called for the http method.
    :param allowed_exceptions: If specified, these exception classes will be
        re-raised instead of turning them into 500 errors.
    :param title: An optional title for the http method.  This will be used
        when generating api documentation.
    :param req_obj_type: A doctor :class:`~doctor.types.Object` type that the
        request body should be converted to.
    """
    def __init__(self, method: str, logic: Callable,
                 allowed_exceptions: List = None, title: str = None,
                 req_obj_type: Callable = None):
        self.method = method
        logic = copy_func(logic)

        # Add doctor attributes to logic.  We do a check to ensure some
        # attributes aren't already set in the event that
        # doctor.utils.add_param_annotations was used to add additional
        # request parameters to the logic function that aren't part of it's
        # signature.
        logic._doctor_req_obj_type = req_obj_type
        if not hasattr(logic, '_doctor_signature'):
            logic._doctor_signature = inspect.signature(logic)
        if not hasattr(logic, '_doctor_params'):
            logic._doctor_params = get_params_from_func(logic)
        logic._doctor_allowed_exceptions = allowed_exceptions
        logic._doctor_title = title
        self.logic = logic


def delete(func: Callable, allowed_exceptions: List = None,
           title: str = None, req_obj_type: Callable = None) -> HTTPMethod:
    """Returns a HTTPMethod instance to create a DELETE route.

    :see: :class:`~doctor.routing.HTTPMethod`
    """
    return HTTPMethod('delete', func, allowed_exceptions=allowed_exceptions,
                      title=title, req_obj_type=req_obj_type)


def get(func: Callable, allowed_exceptions: List = None,
        title: str = None, req_obj_type: Callable = None) -> HTTPMethod:
    """Returns a HTTPMethod instance to create a GET route.

    :see: :class:`~doctor.routing.HTTPMethod`
    """
    return HTTPMethod('get', func, allowed_exceptions=allowed_exceptions,
                      title=title, req_obj_type=req_obj_type)


def post(func: Callable, allowed_exceptions: List = None,
         title: str = None, req_obj_type: Callable = None) -> HTTPMethod:
    """Returns a HTTPMethod instance to create a POST route.

    :see: :class:`~doctor.routing.HTTPMethod`
    """
    return HTTPMethod('post', func, allowed_exceptions=allowed_exceptions,
                      title=title, req_obj_type=req_obj_type)


def put(func: Callable, allowed_exceptions: List = None,
        title: str = None, req_obj_type: Callable = None) -> HTTPMethod:
    """Returns a HTTPMethod instance to create a PUT route.

    :see: :class:`~doctor.routing.HTTPMethod`
    """
    return HTTPMethod('put', func, allowed_exceptions=allowed_exceptions,
                      title=title, req_obj_type=req_obj_type)


def create_http_method(logic: Callable, http_method: str,
                       handle_http: Callable, before: Callable = None,
                       after: Callable = None) -> Callable:
    """Create a handler method to be used in a handler class.

    :param callable logic: The underlying function to execute with the
        parsed and validated parameters.
    :param str http_method: HTTP method this will handle.
    :param handle_http: The HTTP handler function that should be
        used to wrap the logic functions.
    :param before: A function to be called before the logic function associated
        with the route.
    :param after: A function to be called after the logic function associated
        with the route.
    :returns: A handler function.
    """
    @functools.wraps(logic)
    def fn(handler, *args, **kwargs):
        if before is not None and callable(before):
            before()
        result = handle_http(handler, args, kwargs, logic)
        if after is not None and callable(after):
            after(result)
        return result
    return fn


class Route(object):

    """Represents a route.

    :param route: The route path, e.g. `r'^/foo/<int:foo_id>/?$'`
    :param methods: A tuple of defined HTTPMethods for the route.
    :param heading: An optional heading that this route should be grouped
        under in the api documentation.
    :param base_handler_class: The base handler class to use.
    :param handler_name: The name that should be given to the handler class.
    :param before: A function to be called before the logic function associated
        with the route.
    :param after: A function to be called after the logic function associated
        with the route.
    """
    def __init__(self, route: str, methods: Tuple[HTTPMethod],
                 heading: str = 'API', base_handler_class = None,
                 handler_name: str = None, before: Callable = None,
                 after: Callable = None):
        self.after = after
        self.base_handler_class = base_handler_class
        self.before = before
        self.handler_name = handler_name
        self.heading = heading
        self.methods = methods
        self.route = route


def get_handler_name(route: Route, logic: Callable) -> str:
    """Gets the handler name.

    :param route: A Route instance.
    :param logic: The logic function.
    :returns: A handler class name.
    """
    if route.handler_name is not None:
        return route.handler_name
    if any(m for m in route.methods if m.method.lower() == 'post'):
        # A list endpoint
        if route.heading != 'API':
            return '{}ListHandler'.format(get_valid_class_name(route.heading))
        return '{}ListHandler'.format(get_valid_class_name(logic.__name__))
    if route.heading != 'API':
        return '{}Handler'.format(get_valid_class_name(route.heading))
    return '{}Handler'.format(get_valid_class_name(logic.__name__))


def create_routes(routes: Tuple[HTTPMethod], handle_http: Callable,
                  default_base_handler_class: Any) -> List[Tuple[str, Any]]:
    """Creates handler routes from the provided routes.

    :param routes: A tuple containing the route and another tuple with
        all http methods allowed for the route.
    :param handle_http: The HTTP handler function that should be
        used to wrap the logic functions.
    :param default_base_handler_class: The default base handler class that
        should be used.
    :returns: A list of tuples containing the route and generated handler.
    """
    created_routes = []
    all_handler_names = []
    for r in routes:
        handler = None
        if r.base_handler_class is not None:
            base_handler_class = r.base_handler_class
        else:
            base_handler_class = default_base_handler_class

        # Define the handler name.  To prevent issues where auto-generated
        # handler names conflict with existing, appending a number to the
        # end of the hanlder name if it already exists.
        handler_name = get_handler_name(r, r.methods[0].logic)
        if handler_name in all_handler_names:
            handler_name = '{}{}'.format(
                handler_name, len(all_handler_names))
        all_handler_names.append(handler_name)

        for method in r.methods:
            logic = method.logic
            http_method = method.method
            http_func = create_http_method(logic, http_method, handle_http,
                                           before=r.before, after=r.after)

            handler_methods_and_properties = {
                '__name__': handler_name,
                '_doctor_heading': r.heading,
                'methods': set([http_method.upper()]),
                http_method: http_func,
            }
            if handler is None:
                handler = type(
                    handler_name, (base_handler_class,),
                    handler_methods_and_properties)
            else:
                setattr(handler, http_method, http_func)
                # This is specific to Flask.  Its MethodView class
                # initializes the methods attribute in __new__ so we
                # need to add all the other http methods we are defining
                # on the handler after it gets created by type.
                if hasattr(handler, 'methods'):
                    handler.methods.add(http_method.upper())
        created_routes.append((r.route, handler))
    return created_routes



"""
This is a collection of functions used to convert untyped param strings into
their appropriate JSON schema types.
"""

import inspect
import logging
import warnings
from typing import List

import simplejson as json

from doctor.errors import ParseError, TypeSystemError


_bracket_strings = ('[', ord('['))
_brace_strings = ('{', ord('{'))
_false_strings = ('false', b'false')
_true_strings = ('true', b'true')


def _parse_array(value):
    """Coerce value into an list.

    :param str value: Value to parse.
    :returns: list or None if the value is not a JSON array
    :raises: TypeError or ValueError if value appears to be an array but can't
        be parsed as JSON.
    """
    value = value.lstrip()
    if not value or value[0] not in _bracket_strings:
        return None
    return json.loads(value)


def _parse_boolean(value):
    """Coerce value into an bool.

    :param str value: Value to parse.
    :returns: bool or None if the value is not a boolean string.
    """
    value = value.lower()
    if value in _true_strings:
        return True
    elif value in _false_strings:
        return False
    else:
        return None


def _parse_object(value):
    """Coerce value into a dict.

    :param str value: Value to parse.
    :returns: dict or None if the value is not a JSON object
    :raises: TypeError or ValueError if value appears to be an object but can't
        be parsed as JSON.
    """
    value = value.lstrip()
    if not value or value[0] not in _brace_strings:
        return None
    return json.loads(value)


def _parse_string(value):
    """Coerce value into a string.

    This is usually a no-op, but if value is a unicode string, it will be
    encoded as UTF-8 before returning.

    :param str value: Value to parse.
    :returns: str
    """
    if not isinstance(value, str):
        return value.decode('utf-8')
    return value


_parser_funcs = (('boolean', _parse_boolean),
                 ('integer', int),
                 ('number', float),
                 ('array', _parse_array),
                 ('object', _parse_object),
                 ('string', _parse_string))


def parse_value(value, allowed_types, name='value'):
    """Parse a value into one of a number of types.

    This function is used to coerce untyped HTTP parameter strings into an
    appropriate type. It tries to coerce the value into each of the allowed
    types, and uses the first that evaluates properly.

    Because this is coercing a string into multiple, potentially ambiguous,
    types, it tests things in the order of least ambiguous to most ambiguous:

    - The "null" type is checked first. If allowed, and the value is blank
      (""), None will be returned.
    - The "boolean" type is checked next. Values of "true" (case insensitive)
      are True, and values of "false" are False.
    - Numeric types are checked next -- first "integer", then "number".
    - The "array" type is checked next. A value is only considered a valid
      array if it begins with a "[" and can be parsed as JSON.
    - The "object" type is checked next. A value is only considered a valid
      object if it begins with a "{" and can be parsed as JSON.
    - The "string" type is checked last, since any value is a valid string.
      Unicode strings are encoded as UTF-8.

    :param str value: Parameter value. Example: "1"
    :param list allowed_types: Types that should be attempted. Example:
        ["integer", "null"]
    :param str name: Parameter name. If not specified, "value" is used.
        Example: "campaign_id"
    :returns: a tuple of a type string and coerced value
    :raises: ParseError if the value cannot be coerced to any of the types
    """
    if not isinstance(value, str):
        raise ValueError('value for %r must be a string' % name)
    if isinstance(allowed_types, str):
        allowed_types = [allowed_types]

    # Note that the order of these type considerations is important. Because we
    # have an untyped value that may be one of any given number of types, we
    # need a consistent order of evaluation in cases when there is ambiguity
    # between types.

    if 'null' in allowed_types and value == '':
        return 'null', None

    # For all of these types, we'll pass the value to the function and it will
    # raise a TypeError or ValueError or return None if it can't be parsed as
    # the given type.

    for allowed_type, parser in _parser_funcs:
        if allowed_type in allowed_types:
            try:
                parsed_value = parser(value)
                if parsed_value is not None:
                    return allowed_type, parsed_value
            except (TypeError, ValueError):
                # Ignore any errors, and continue trying other types
                pass

    raise ParseError('%s must be a valid type (%s)' %
                     (name, ', '.join(allowed_types)))


def parse_json(value: str, sig_params: List[inspect.Parameter] = None) -> dict:
    """Parse a value as JSON.

    This is just a wrapper around json.loads which re-raises any errors as a
    ParseError instead.

    :param str value: JSON string.
    :param dict sig_params: The logic function's signature parameters.
    :returns: the parsed JSON value
    """
    try:
        loaded = json.loads(value)
    except Exception as e:
        message = 'Error parsing JSON: %r error: %s' % (value, e)
        logging.debug(message, exc_info=e)
        raise ParseError(message)

    if sig_params is not None:
        return map_param_names(loaded, sig_params)
    return loaded


_native_type_to_json = {
    list: 'array',
    bool: 'boolean',
    int: 'integer',
    dict: 'object',
    float: 'number',
    str: 'string'
}


def map_param_names(
        req_params: dict, sig_params: List[inspect.Parameter]) -> dict:
    """Maps request param names to match logic function param names.

    If a doctor type defined a `param_name` attribute for the name of the
    parameter in the request, we should use that as the key when looking up
    the value for the request parameter.

    When we declare a type we can specify what the parameter name
    should be in the request that the annotated type should get mapped to.

    >>> from doctor.types import number
    >>> Latitude = number('The latitude', param_name='location.lat')
    >>> def my_logic(lat: Latitude): pass
    >>> request_params = {'location.lat': 45.2342343}

    In the above example doctor knows to pass the value at key `location.lat`
    to the logic function variable named `lat` since it's annotated by the
    `Latitude` type which specifies what the param_name is on the request.

    :param dict req_params: The parameters specified in the request.
    :param dict sig_params: The logic function's signature parameters.
    :returns: A dict of re-mapped params.
    """
    new_request_params = {}
    for k, param in sig_params.items():
        param_name = getattr(param.annotation, 'param_name', None)
        key = k if param_name is None else param_name
        if key in req_params:
            new_request_params[k] = req_params[key]
    return new_request_params


def parse_form_and_query_params(req_params: dict, sig_params: dict) -> dict:
    """Uses the parameter annotations to coerce string params.

    This is used for HTTP requests, in which the form parameters are all
    strings, but need to be converted to the appropriate types before
    validating them.

    :param dict req_params: The parameters specified in the request.
    :param dict sig_params: The logic function's signature parameters.
    :returns: a dict of params parsed from the input dict.
    :raises TypeSystemError: If there are errors parsing values.
    """
    # Importing here to prevent circular dependencies.
    from doctor.types import SuperType, UnionType
    errors = {}
    parsed_params = {}
    for param, value in req_params.items():
        # Skip request variables not in the function signature.
        if param not in sig_params:
            continue
        # Skip coercing parameters not annotated by a doctor type.
        if not issubclass(sig_params[param].annotation, SuperType):
            continue

        # Check if the type has a custom parser for the parameter.
        custom_parser = sig_params[param].annotation.parser
        if custom_parser is not None:
            if not callable(custom_parser):
                warnings.warn(
                    'Parser `{}` is not callable, using default parser.'.format(
                        custom_parser))
                custom_parser = None

        try:
            if custom_parser is not None:
                parsed_params[param] = custom_parser(value)
            else:
                if issubclass(sig_params[param].annotation, UnionType):
                    json_type = [
                        _native_type_to_json[_type.native_type]
                        for _type in sig_params[param].annotation.types
                    ]
                else:
                    native_type = sig_params[param].annotation.native_type
                    json_type = [_native_type_to_json[native_type]]
                # If the type is nullable, also add null as an allowed type.
                if sig_params[param].annotation.nullable:
                    json_type.append('null')
                _, parsed_params[param] = parse_value(value, json_type)
        except ParseError as e:
            errors[param] = str(e)

    if errors:
        raise TypeSystemError(errors, errors=errors)

    return parsed_params



# encoding: utf-8

import inspect
import json

import pytest

from doctor.errors import ParseError, TypeSystemError
from doctor.parsers import (
    map_param_names, parse_form_and_query_params, parse_json, parse_value,
    _parse_string)
from doctor.types import string

from .base import TestCase
from .types import (
    Age, Auth, Color, ColorsOrObject, FoosWithParser, IsDeleted, Latitude,
    Longitude, OptIn)


def logic(age: Age, color: Color, is_deleted: IsDeleted=False):
    return '{} {} {}'.format(age, color, is_deleted)


def logic2(foos: FoosWithParser):
    return foos


def logic_with_union(colors_or_object: ColorsOrObject):
    return colors_or_object


class TestParsers(TestCase):

    def test_parse_json(self):
        assert {'foo': 1} == parse_json('{"foo": 1}')
        message = 'Error parsing JSON: \'bad json\' error: Expecting value'
        with pytest.raises(ParseError, match=message):
            parse_json('bad json')

    def test_parse_json_with_sig_params(self):
        """
        Verifies if we pass a signature it maps parameters properly.
        """
        def func(lat: Latitude):
            pass

        sig = inspect.signature(func)
        request_params = json.dumps({
            'location.lat': 127.11,
        })
        actual = parse_json(request_params, sig.parameters)
        assert {'lat': 127.11} == actual

    def test_parse_value_allowed_types_is_str(self):
        assert ('integer', 12) == parse_value('12', allowed_types='integer')

    def test_parse_value(self):
        """These cases should return a value when parsing."""
        tests = [
            ('', ['null'], None),
            ('true', ['boolean'], True),
            ('True', ['boolean'], True),
            ('TRUE', ['boolean'], True),
            ('false', ['boolean'], False),
            ('False', ['boolean'], False),
            ('FALSE', ['boolean'], False),
            ('0', ['integer'], 0),
            ('1', ['integer'], 1),
            ('-1', ['integer'], -1),
            ('0', ['number'], 0.0),
            ('0.5', ['number'], 0.5),
            ('[]', ['array'], []),
            ('[0]', ['array'], [0]),
            ('["0", 0]', ['array'], ['0', 0]),
            ('{}', ['object'], {}),
            ('{"a": 1}', ['object'], {'a': 1}),
            ('foo', ['string'], 'foo'),
            ('"foo"', ['string'], '"foo"'),
            (u'foø', ['string'], u'foø'),

            # ambiguous test cases
            ('', ['string', 'null'], None),
            ('', ['boolean', 'string'], ''),
            ('', ['boolean', 'string', 'null'], None),
            ('true', ['string', 'boolean'], True),
            ('false', ['string', 'boolean'], False),
            ('0', ['string', 'float', 'integer'], 0),
            ('1', ['string', 'float', 'integer'], 1),
            ('[]', ['string', 'object', 'array'], []),
            ('  []', ['string', 'object', 'array'], []),
            ('{}', ['string', 'array', 'object'], {}),
            ('    {}', ['string', 'array', 'object'], {}),
        ]
        for value, allowed_types, expected_value in tests:
            expected = (allowed_types[-1], expected_value)
            assert parse_value(value, allowed_types) == expected

    def test_parse_value_errors(self):
        """These cases should raise an error when parsing."""
        tests = {'array': ['{}', '['],
                 'boolean': ['0', '1'],
                 'float': ['bad'],
                 'integer': ['bad', '0.1'],
                 'object': ['[]', '{']}
        for allowed_type, bad_values in list(tests.items()):
            for bad_value in bad_values:
                with pytest.raises(ParseError):
                    parse_value(bad_value, [allowed_type])

    def test_parse_value_error_not_string(self):
        """Should raise an error if the value isn't a string."""
        with pytest.raises(
                ValueError, match=r"value for 'foo' must be a string"):
            parse_value(1337, [], 'foo')

    def test_parse_string_not_str(self):
        actual = _parse_string(b'foo')
        assert 'foo' == actual

    def test_parse_form_and_query_params_with_union_type(self):
        """
        This test verifies if one of the annotations is a UnionType, that we
        correctly parse the value as it can be one of many types.
        """
        sig = inspect.signature(logic_with_union)
        query_params = {'colors_or_object': '["blue"]'}
        actual = parse_form_and_query_params(query_params, sig.parameters)
        assert {'colors_or_object': ['blue']} == actual

        query_params = {'colors_or_object': json.dumps({'str': 'auth'})}
        actual = parse_form_and_query_params(query_params, sig.parameters)
        expected = {
            'colors_or_object': {'str': 'auth'},
        }
        assert expected == actual

        # An invalid type that isn't one of the accepted types
        query_params = {'colors_or_object': 45}
        with pytest.raises(ValueError):
            parse_form_and_query_params(query_params, sig.parameters)

    def test_parse_form_and_query_params_no_errors_with_custom_parser(self):
        sig = inspect.signature(logic2)
        query_params = {
            'foos': 'one,two,three',
        }
        actual = parse_form_and_query_params(query_params, sig.parameters)
        expected = {
            'foos': ['one', 'two', 'three'],
        }
        assert expected == actual

    def test_parse_form_and_query_params_custom_parser_for_some_params(self):
        """
        This test just verifies that we can have types with custom parser mixed
        with those that use the default parsers.
        """
        def f(age: Age, items: FoosWithParser, color: Color):
            pass

        sig = inspect.signature(f)
        query_params = {
            'age': '22',
            'color': 'green',
            'items': 'item1,item2',
        }
        actual = parse_form_and_query_params(query_params, sig.parameters)
        expected = {
            'age': 22,
            'color': 'green',
            'items': ['item1', 'item2'],
        }
        assert expected == actual

    def test_parse_form_and_query_params_with_custom_parser_not_callable(self):
        """
        This test verifies if a parser is provided that isn't callable that
        we warn the user and fallback to the default parser.
        """
        A = string('str', parser='foo')

        def f(a: A):
            pass

        sig = inspect.signature(f)
        query_params = {'a': 'a'}
        with pytest.warns(UserWarning, match='Parser `foo` is not callable'):
            actual = parse_form_and_query_params(query_params, sig.parameters)
        assert {'a': 'a'} == actual

    def test_parse_form_and_query_params_no_errors(self):
        sig = inspect.signature(logic)
        query_params = {
            'age': '22',
            'color': 'blue',
            'is_deleted': 'true',
        }
        actual = parse_form_and_query_params(query_params, sig.parameters)
        assert {
            'age': 22,
            'color': 'blue',
            'is_deleted': True,
        } == actual

    def test_parse_form_and_query_params_with_errors(self):
        sig = inspect.signature(logic)
        query_params = {
            'age': 'not an int',
            'color': 'blue',
            'is_deleted': 'not a boolean',
        }
        with pytest.raises(TypeSystemError) as exc:
            parse_form_and_query_params(query_params, sig.parameters)

        assert {
            'age': 'value must be a valid type (integer)',
            'is_deleted': 'value must be a valid type (boolean)',
        } == exc.value.errors

    def test_parse_form_and_query_params_no_doctor_type_param_in_sig(self):
        """
        This is a regression test for when a logic function has a parameter
        in it's signature that is not annotated by a doctor type and the name
        of the parameter matches a parameter in the request variables.
        Previously this would cause an AttributeError.
        """
        def logic(age: Age, use_cache: bool = False):
            pass

        sig = inspect.signature(logic)
        params = {
            'age': '22',
            'use_cache': '1',
        }
        actual = parse_form_and_query_params(params, sig.parameters)
        expected = {'age': 22}
        assert expected == actual

    def test_map_param_names(seilf):
        def foo(lat: Latitude, lon: Longitude, opt_in: OptIn, auth: Auth):
            pass

        sig = inspect.signature(foo)
        request_params = {
            'auth': 'auth',
            'location.lat': 45.12345,
            'locationLon': -122.12345,
            'opt-in': True,
        }
        actual = map_param_names(request_params, sig.parameters)

        expected = {
            'auth': 'auth',
            'lat': 45.12345,
            'lon': -122.12345,
            'opt_in': True,
        }
        assert expected == actual



"""
Copyright © 2017, Encode OSS Ltd. All rights reserved.

Redistribution and use in source and binary forms, with or without modification,
are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

Redistributions in binary form must reproduce the above copyright notice, this
list of conditions and the following disclaimer in the documentation and/or
other materials provided with the distribution.

Neither the name of the copyright holder nor the names of its contributors may
be used to endorse or promote products derived from this software without
specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

This file is a modified version of the typingsystem.py module in apistar.
https://github.com/encode/apistar/blob/973c6485d8297c1bcef35a42221ac5107dce25d5/apistar/typesystem.py
"""
import math
import re
import typing
from datetime import datetime

import isodate
import rfc3987

from doctor.errors import SchemaError, SchemaValidationError, TypeSystemError
from doctor.parsers import parse_value


StrOrList = typing.Union[str, typing.List[str]]


class classproperty(object):
    """A decorator that allows a class to contain a class property.

    This is a function that can be executed on a non-instance but accessed
    via a property.

    >>> class Foo(object):
    ...   a = 1
    ...   @classproperty
    ...   def b(cls):
    ...     return cls.a + 1
    ...
    >>> Foo.b
    2
    """

    def __init__(self, fget):
        self.fget = fget

    def __get__(self, owner_self, owner_cls):
        return self.fget(owner_cls)


class MissingDescriptionError(ValueError):
    """An exception raised when a type is missing a description."""
    pass


class SuperType(object):
    """A super type all custom types must extend from.

    This super type requires all subclasses define a description attribute
    that describes what the type represents.  A `ValueError` will be raised
    if the subclass does not define a `description` attribute.
    """
    #: The description of what the type represents.
    description = None  # type: str

    #: An example value for the type.
    example = None

    #: Indicates if the value of this type is allowed to be None.
    nullable = False  # type: bool

    #: An optional name of where to find the request parameter if it does not
    #: match the variable name in your logic function.
    param_name = None  # type: str

    #: An optional callable to parse a request paramter before it gets validated
    #: by a type.  It should accept a single value paramter and return the
    #: parsed value.
    parser = None  # type: typing.Callable

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.description is None:
            cls = self.__class__
            raise MissingDescriptionError(
                '{} did not define a description attribute'.format(cls))

    @classmethod
    def validate(cls, value: typing.Any):
        """Additional validation for a type.

        All types will have a validate method where custom validation logic
        can be placed.  The implementor should return nothing if the value is
        valid, otherwise a `TypeSystemError` should be raised.

        :param value: The value to be validated.
        """
        pass


class UnionType(SuperType):
    """A type that can be one of any of the defined `types`.

    The first type that does not raise a :class:`~doctor.errors.TypeSystemError`
    will be used as the type for the variable.
    """
    #: A list of allowed types.
    types = []

    _native_type = None

    def __new__(cls, *args, **kwargs):
        if not cls.types:
            raise TypeSystemError(
                'Sub-class must define a `types` list attribute containing at '
                'least 1 type.', cls=cls)

        valid = False
        value = None
        errors = {}
        for obj_class in cls.types:
            try:
                value = obj_class(*args, **kwargs)
                valid = True
                # Dynamically change the native_type based on that of the value.
                cls._native_type = obj_class.native_type
                break
            except TypeSystemError as e:
                errors[obj_class.__name__] = str(e)
                continue

        if not valid:
            klasses = [klass.__name__ for klass in cls.types]
            raise TypeSystemError('Value is not one of {}. {}'.format(
                klasses, errors))

        cls.validate(value)
        return value

    @classmethod
    def get_example(cls):
        """Returns an example value for the UnionType."""
        return cls.types[0].get_example()

    @classproperty
    def native_type(cls):
        """Returns the native type.

        Since UnionType can have multiple types, simply return the native type
        of the first type defined in the types attribute.

        If _native_type is set based on initializing a value with the class,
        then we return the dynamically modified type that matches that of the
        value used during instantiation.  e.g.

        >>> from doctor.types import UnionType, string, boolean
        >>> class BoolOrStr(UnionType):
        ...   description = 'bool or str'
        ...   types = [boolean('a bool'), string('a string')]
        ...
        >>> BoolOrStr.native_type
        <class 'bool'>
        >>> BoolOrStr('str')
        'str'
        >>> BoolOrStr.native_type
        <class 'str'>
        >>> BoolOrStr(False)
        False
        >>> BoolOrStr.native_type
        <class 'bool'>
        """
        if cls._native_type is not None:
            return cls._native_type
        return cls.types[0].native_type


class String(SuperType, str):
    """Represents a `str` type."""
    native_type = str
    errors = {
        'blank': 'Must not be blank.',
        'max_length': 'Must have no more than {max_length} characters.',
        'min_length': 'Must have at least {min_length} characters.',
        'pattern': 'Must match the pattern /{pattern}/.',
    }
    #: Will check format of the string for `date`, `date-time`, `email`,
    #: `time` and `uri`.
    format = None
    #: The maximum length of the string.
    max_length = None  # type: int
    #: The minimum length of the string.
    min_length = None  # type: int
    #: A regex pattern that the string should match.
    pattern = None  # type: str
    #: Whether to trim whitespace on a string.  Defaults to `True`.
    trim_whitespace = True

    def __new__(cls, *args, **kwargs):
        if cls.nullable and args[0] is None:
            return None

        value = super().__new__(cls, *args, **kwargs)

        if cls.trim_whitespace:
            value = value.strip()

        if cls.min_length is not None:
            if len(value) < cls.min_length:
                if cls.min_length == 1:
                    raise TypeSystemError(cls=cls, code='blank')
                else:
                    raise TypeSystemError(cls=cls, code='min_length')

        if cls.max_length is not None:
            if len(value) > cls.max_length:
                raise TypeSystemError(cls=cls, code='max_length')

        if cls.pattern is not None:
            if not re.search(cls.pattern, value):
                raise TypeSystemError(cls=cls, code='pattern')

        # Validate format, if specified
        if cls.format == 'date':
            try:
                value = datetime.strptime(value, "%Y-%m-%d").date()
            except ValueError as e:
                raise TypeSystemError(str(e), cls=cls)
        elif cls.format == 'date-time':
            try:
                value = isodate.parse_datetime(value)
            except (ValueError, isodate.ISO8601Error) as e:
                raise TypeSystemError(str(e), cls=cls)
        elif cls.format == 'email':
            if '@' not in value:
                raise TypeSystemError('Not a valid email address.', cls=cls)
        elif cls.format == 'time':
            try:
                value = datetime.strptime(value, "%H:%M:%S")
            except ValueError as e:
                raise TypeSystemError(str(e), cls=cls)
        elif cls.format == 'uri':
            try:
                rfc3987.parse(value, rule='URI')
            except ValueError as e:
                raise TypeSystemError(str(e), cls=cls)

        # Coerce value to the native str type.  We only do this if the value
        # is an instance of the class.  It could be a datetime instance or
        # a str already if `trim_whitespace` is True.
        if isinstance(value, cls):
            value = cls.native_type(value)

        cls.validate(value)
        return value

    @classmethod
    def get_example(cls) -> str:
        """Returns an example value for the String type."""
        if cls.example is not None:
            return cls.example
        return 'string'


class _NumericType(SuperType):
    """
    Base class for both `Number` and `Integer`.
    """
    native_type = None  # type: type
    errors = {
        'type': 'Must be a valid number.',
        'finite': 'Must be a finite number.',
        'minimum': 'Must be greater than or equal to {minimum}.',
        'exclusive_minimum': 'Must be greater than {minimum}.',
        'maximum': 'Must be less than or equal to {maximum}.',
        'exclusive_maximum': 'Must be less than {maximum}.',
        'multiple_of': 'Must be a multiple of {multiple_of}.',
    }
    #: The minimum value allowed.
    minimum = None  # type: typing.Union[float, int]
    #: The maximum value allowed.
    maximum = None  # type: typing.Union[float, int]
    #: The minimum value should be treated as exclusive or not.
    exclusive_minimum = False
    #: The maximum value should be treated as exclusive or not.
    exclusive_maximum = False
    #: The value is required to be a multiple of this value.
    multiple_of = None  # type: typing.Union[float, int]

    def __new__(cls, *args, **kwargs):
        if cls.nullable and args[0] is None:
            return None

        try:
            value = cls.native_type.__new__(cls, *args, **kwargs)
        except (TypeError, ValueError):
            raise TypeSystemError(cls=cls, code='type') from None

        if not math.isfinite(value):
            raise TypeSystemError(cls=cls, code='finite')

        if cls.minimum is not None:
            if cls.exclusive_minimum:
                if value <= cls.minimum:
                    raise TypeSystemError(cls=cls, code='exclusive_minimum')
            else:
                if value < cls.minimum:
                    raise TypeSystemError(cls=cls, code='minimum')

        if cls.maximum is not None:
            if cls.exclusive_maximum:
                if value >= cls.maximum:
                    raise TypeSystemError(cls=cls, code='exclusive_maximum')
            else:
                if value > cls.maximum:
                    raise TypeSystemError(cls=cls, code='maximum')

        if cls.multiple_of is not None:
            if isinstance(cls.multiple_of, float):
                failed = not (value * (1 / cls.multiple_of)).is_integer()
            else:
                failed = value % cls.multiple_of
            if failed:
                raise TypeSystemError(cls=cls, code='multiple_of')

        # Coerce value to the native type.  We only do this if the value
        # is an instance of the class.
        if isinstance(value, cls):
            value = cls.native_type(value)

        cls.validate(value)
        return value


class Number(_NumericType, float):
    """Represents a `float` type."""
    native_type = float

    @classmethod
    def get_example(cls) -> float:
        """Returns an example value for the Number type."""
        if cls.example is not None:
            return cls.example
        return 3.14


class Integer(_NumericType, int):
    """Represents an `int` type."""
    native_type = int

    @classmethod
    def get_example(cls) -> int:
        """Returns an example value for the Integer type."""
        if cls.example is not None:
            return cls.example
        return 1


class Boolean(SuperType):
    """Represents a `bool` type."""
    native_type = bool
    errors = {
        'type': 'Must be a valid boolean.'
    }

    def __new__(cls, *args, **kwargs) -> bool:
        value = args[0]
        if cls.nullable and value is None:
            return None

        if args and isinstance(value, str):
            try:
                value = {
                    'true': True,
                    'false': False,
                    'on': True,
                    'off': False,
                    '1': True,
                    '0': False,
                    '': False
                }[value.lower()]
            except KeyError:
                raise TypeSystemError(cls=cls, code='type') from None
            cls.validate(value)
            return value

        cls.validate(value)
        return bool(*args, **kwargs)

    @classmethod
    def get_example(cls) -> bool:
        """Returns an example value for the Boolean type."""
        if cls.example is not None:
            return cls.example
        return True


class Enum(SuperType, str):
    """
    Represents a `str` type that must be one of any defined allowed values.
    """
    native_type = str
    errors = {
        'invalid': 'Must be one of: {enum}',
    }
    #: A list of valid values.
    enum = []  # type: typing.List[str]

    #: Indicates if the values of the enum are case insensitive or not.
    case_insensitive = False

    #: If True the input value will be lowercased before validation.
    lowercase_value = False

    #: If True the input value will be uppercased before validation.
    uppercase_value = False

    def __new__(cls, value: typing.Union[None, str]):
        if cls.nullable and value is None:
            return None

        if cls.case_insensitive:
            if cls.uppercase_value:
                cls.enum = [v.upper() for v in cls.enum]
            else:
                cls.enum = [v.lower() for v in cls.enum]
                value = value.lower()
        if cls.lowercase_value:
            value = value.lower()
        if cls.uppercase_value:
            value = value.upper()
        if value not in cls.enum:
            raise TypeSystemError(cls=cls, code='invalid')

        cls.validate(value)
        return value

    @classmethod
    def get_example(cls) -> str:
        """Returns an example value for the Enum type."""
        if cls.example is not None:
            return cls.example
        return cls.enum[0]


class Object(SuperType, dict):
    """Represents a `dict` type."""
    native_type = dict
    errors = {
        'type': 'Must be an object.',
        'invalid_key': 'Object keys must be strings.',
        'required': 'This field is required.',
        'additional_properties': 'Additional properties are not allowed.',
    }
    #: A mapping of property name to expected type.
    properties = {}  # type: typing.Dict[str, typing.Any]
    #: A list of required properties.
    required = []  # type: typing.List[str]
    #: If True additional properties will be allowed, otherwise they will not.
    additional_properties = True  # type: bool
    #: A human readable title for the object.
    title = None
    #: A mapping of property name to a list of other properties it requires
    #: when the property name is present.
    property_dependencies = {}  # type: typing.Dict[str, typing.List[str]]

    def __init__(self, *args, **kwargs):
        if self.nullable and args[0] is None:
            return

        try:
            super().__init__(*args, **kwargs)
        except MissingDescriptionError:
            raise
        except (ValueError, TypeError):
            if (len(args) == 1 and not kwargs and
                    hasattr(args[0], '__dict__')):
                value = dict(args[0].__dict__)
            else:
                raise TypeSystemError(
                    cls=self.__class__, code='type') from None
        value = self

        # Ensure all property keys are strings.
        errors = {}
        if any(not isinstance(key, str) for key in value.keys()):
            raise TypeSystemError(cls=self.__class__, code='invalid_key')

        # Properties
        for key, child_schema in self.properties.items():
            try:
                item = value[key]
            except KeyError:
                if hasattr(child_schema, 'default'):
                    # If a key is missing but has a default, then use that.
                    self[key] = child_schema.default
                elif key in self.required:
                    exc = TypeSystemError(cls=self.__class__, code='required')
                    errors[key] = exc.detail
            else:
                # Coerce value into the given schema type if needed.
                if isinstance(item, child_schema):
                    self[key] = item
                else:
                    try:
                        self[key] = child_schema(item)
                    except TypeSystemError as exc:
                        errors[key] = exc.detail

        # If additional properties are allowed set any other key/value(s) not
        # in the defined properties.
        if self.additional_properties:
            for key, value in value.items():
                if key not in self:
                    self[key] = value

        # Raise an exception if additional properties are defined and
        # not allowed.
        if not self.additional_properties:
            properties = list(self.properties.keys())
            for key in self.keys():
                if key not in properties:
                    detail = '{key} not in {properties}'.format(
                        key=key, properties=properties)
                    exc = TypeSystemError(detail, cls=self.__class__,
                                          code='additional_properties')
                    errors[key] = exc.detail

        # Check for any property dependencies that are defined.
        if self.property_dependencies:
            err = 'Required properties {} for property `{}` are missing.'
            for prop, dependencies in self.property_dependencies.items():
                if prop in self:
                    for dep in dependencies:
                        if dep not in self:
                            raise TypeSystemError(err.format(
                                dependencies, prop))

        if errors:
            raise TypeSystemError(errors)

        self.validate(self.copy())

    @classmethod
    def get_example(cls) -> dict:
        """Returns an example value for the Dict type.

        If an example isn't a defined attribute on the class we return
        a dict of example values based on each property's annotation.
        """
        if cls.example is not None:
            return cls.example
        return {k: v.get_example() for k, v in cls.properties.items()}


class Array(SuperType, list):
    """Represents a `list` type."""
    native_type = list
    errors = {
        'type': 'Must be a list.',
        'min_items': 'Not enough items.',
        'max_items': 'Too many items.',
        'unique_items': 'This item is not unique.',
    }
    #: The type each item should be, or a list of types where the position
    #: of the type in the list represents the type at that position in the
    #: array the item should be.
    items = None  # type: typing.Union[type, typing.List[type]]
    #: If `items` is a list and this is `True` then additional items whose
    #: types aren't defined are allowed in the list.
    additional_items = False  # type: bool
    #: The minimum number of items allowed in the list.
    min_items = 0  # type: typing.Optional[int]
    #: The maxiimum number of items allowed in the list.
    max_items = None  # type: typing.Optional[int]
    #: If `True` items in the array should be unique from one another.
    unique_items = False  # type: bool

    def __init__(self, *args, **kwargs):
        if self.nullable and args[0] is None:
            return

        if args and isinstance(args[0], (str, bytes)):
            raise TypeSystemError(cls=self.__class__, code='type')

        try:
            value = list(*args, **kwargs)
        except TypeError:
            raise TypeSystemError(cls=self.__class__, code='type') from None

        if isinstance(self.items, list) and len(self.items) > 1:
            if len(value) < len(self.items):
                raise TypeSystemError(cls=self.__class__, code='min_items')
            elif len(value) > len(self.items) and not self.additional_items:
                raise TypeSystemError(cls=self.__class__, code='max_items')

        if len(value) < self.min_items:
            raise TypeSystemError(cls=self.__class__, code='min_items')
        elif self.max_items is not None and len(value) > self.max_items:
            raise TypeSystemError(cls=self.__class__, code='max_items')

        # Ensure all items are of the right type.
        errors = {}
        if self.unique_items:
            seen_items = set()

        for pos, item in enumerate(value):
            try:
                if isinstance(self.items, list):
                    if pos < len(self.items):
                        item = self.items[pos](item)
                elif self.items is not None:
                    item = self.items(item)

                if self.unique_items:
                    if item in seen_items:
                        raise TypeSystemError(
                            cls=self.__class__, code='unique_items')
                    else:
                        seen_items.add(item)

                self.append(item)
            except TypeSystemError as exc:
                errors[pos] = exc.detail

        if errors:
            raise TypeSystemError(errors)

        self.validate(value)

    @classmethod
    def get_example(cls) -> list:
        """Returns an example value for the Array type.

        If an example isn't a defined attribute on the class we return
        a list of 1 item containing the example value of the `items` attribute.
        If `items` is None we simply return a `[1]`.
        """
        if cls.example is not None:
            return cls.example
        if cls.items is not None:
            if isinstance(cls.items, list):
                return [item.get_example() for item in cls.items]
            else:
                return [cls.items.get_example()]
        return [1]


class JsonSchema(SuperType):
    """Represents a type loaded from a json schema.

    NOTE: This class should not be used directly.  Instead use
    :func:`~doctor.types.json_schema_type` to create a new class based on
    this one.
    """
    json_type = None
    native_type = None
    #: The loaded ResourceSchema
    schema = None  # type: doctor.resource.ResourceSchema
    #: The full path to the schema file.
    schema_file = None  # type: str
    #: The key from the definitions in the schema file that the type should
    #: come from.
    definition_key = None  # type: str

    def __new__(cls, value):
        # Attempt to parse the value if it came from a query string
        try:
            _, value = parse_value(value, [cls.json_type])
        except ValueError:
            pass
        request_schema = None
        if cls.definition_key is not None:
            params = [cls.definition_key]
            request_schema = cls.schema._create_request_schema(params, params)
            data = {cls.definition_key: value}
        else:
            data = value

        super().__new__(cls)
        # Validate the data against the schema and raise an error if it
        # does not validate.
        validator = cls.schema.get_validator(request_schema)
        try:
            cls.schema.validate(data, validator)
        except SchemaValidationError as e:
            raise TypeSystemError(e.args[0], cls=cls)

        return value

    @classmethod
    def get_example(cls) -> typing.Any:
        """Returns an example value for the JsonSchema type."""
        return cls.example


#: A mapping of json types to native python types.
JSON_TYPES_TO_NATIVE = {
    'array': list,
    'boolean': bool,
    'integer': int,
    'object': dict,
    'number': float,
    'string': str,
}


def get_value_from_schema(schema, definition: dict, key: str,
                          definition_key: str):
    """Gets a value from a schema and definition.

    If the value has references it will recursively attempt to resolve them.

    :param ResourceSchema schema: The resource schema.
    :param dict definition: The definition dict from the schema.
    :param str key: The key to use to get the value from the schema.
    :param str definition_key: The name of the definition.
    :returns: The value.
    :raises TypeSystemError: If the key can't be found in the schema/definition
        or we can't resolve the definition.
    """
    resolved_definition = definition.copy()
    if '$ref' in resolved_definition:
        try:
            # NOTE: The resolve method recursively resolves references, so
            # we don't need to worry about that in this function.
            resolved_definition = schema.resolve(definition['$ref'])
        except SchemaError as e:
            raise TypeSystemError(str(e))
    try:
        value = resolved_definition[key]
    except KeyError:
        # Before raising an error, the resolved definition may have an array
        # or object inside it that needs to be resolved in order to get
        # values.  Attempt that here and then fail if we still can't find
        # the key we are looking for.

        # If the key was missing and this is an array, try to resolve it
        # from the items key.
        if resolved_definition['type'] == 'array':
            return [
                get_value_from_schema(schema, resolved_definition['items'], key,
                                      definition_key)
            ]
        # If the key was missing and this is an object, resolve it from it's
        # properties.
        elif resolved_definition['type'] == 'object':
            value = {}
            for prop, definition in resolved_definition['properties'].items():
                value[prop] = get_value_from_schema(
                    schema, definition, key, definition_key)
            return value
        raise TypeSystemError(
            'Definition `{}` is missing a {}.'.format(
                definition_key, key))
    return value


def get_types(json_type: StrOrList) -> typing.Tuple[str, str]:
    """Returns the json and native python type based on the json_type input.

    If json_type is a list of types it will return the first non 'null' value.

    :param json_type: A json type or a list of json types.
    :returns: A tuple containing the json type and native python type.
    """
    # If the type is a list, use the first non 'null' value as the type.
    if isinstance(json_type, list):
        for j_type in json_type:
            if j_type != 'null':
                json_type = j_type
                break
    return (json_type, JSON_TYPES_TO_NATIVE[json_type])


def json_schema_type(schema_file: str, **kwargs) -> typing.Type:
    """Create a :class:`~doctor.types.JsonSchema` type.

    This function will automatically load the schema and set it as an attribute
    of the class along with the description and example.

    :param schema_file: The full path to the json schema file to load.
    :param kwargs: Can include any attribute defined in
        :class:`~doctor.types.JsonSchema`
    """
    # Importing here to avoid circular dependencies
    from doctor.resource import ResourceSchema
    schema = ResourceSchema.from_file(schema_file)
    kwargs['schema'] = schema

    # Look up the description, example and type in the schema.
    definition_key = kwargs.get('definition_key')
    if definition_key:
        params = [definition_key]
        request_schema = schema._create_request_schema(params, params)
        try:
            definition = request_schema['definitions'][definition_key]
        except KeyError:
            raise TypeSystemError(
                'Definition `{}` is not defined in the schema.'.format(
                    definition_key))
        description = get_value_from_schema(
            schema, definition, 'description', definition_key)
        example = get_value_from_schema(
            schema, definition, 'example', definition_key)
        json_type = get_value_from_schema(
            schema, definition, 'type', definition_key)
        json_type, native_type = get_types(json_type)
        kwargs['description'] = description
        kwargs['example'] = example
        kwargs['json_type'] = json_type
        kwargs['native_type'] = native_type
    else:
        try:
            kwargs['description'] = schema.schema['description']
        except KeyError:
            raise TypeSystemError('Schema is missing a description.')
        try:
            json_type = schema.schema['type']
        except KeyError:
            raise TypeSystemError('Schema is missing a type.')
        json_type, native_type = get_types(json_type)
        kwargs['json_type'] = json_type
        kwargs['native_type'] = native_type
        try:
            kwargs['example'] = schema.schema['example']
        except KeyError:
            # Attempt to load from properties, if defined.
            if schema.schema.get('properties'):
                example = {}
                for prop, definition in schema.schema['properties'].items():
                    example[prop] = get_value_from_schema(
                        schema, definition, 'example', 'root')
                kwargs['example'] = example
            else:
                raise TypeSystemError('Schema is missing an example.')

    return type('JsonSchema', (JsonSchema,), kwargs)


def string(description: str, **kwargs) -> typing.Type:
    """Create a :class:`~doctor.types.String` type.

    :param description: A description of the type.
    :param kwargs: Can include any attribute defined in
        :class:`~doctor.types.String`
    """
    kwargs['description'] = description
    return type('String', (String,), kwargs)


def integer(description, **kwargs) -> typing.Type:
    """Create a :class:`~doctor.types.Integer` type.

    :param description: A description of the type.
    :param kwargs: Can include any attribute defined in
        :class:`~doctor.types.Integer`
    """
    kwargs['description'] = description
    return type('Integer', (Integer,), kwargs)


def number(description, **kwargs) -> typing.Type:
    """Create a :class:`~doctor.types.Number` type.

    :param description: A description of the type.
    :param kwargs: Can include any attribute defined in
        :class:`~doctor.types.Number`
    """
    kwargs['description'] = description
    return type('Number', (Number,), kwargs)


def boolean(description, **kwargs) -> typing.Type:
    """Create a :class:`~doctor.types.Boolean` type.

    :param description: A description of the type.
    :param kwargs: Can include any attribute defined in
        :class:`~doctor.types.Boolean`
    """
    kwargs['description'] = description
    return type('Boolean', (Boolean,), kwargs)


def enum(description, **kwargs) -> typing.Type:
    """Create a :class:`~doctor.types.Enum` type.

    :param description: A description of the type.
    :param kwargs: Can include any attribute defined in
        :class:`~doctor.types.Enum`
    """
    kwargs['description'] = description
    return type('Enum', (Enum,), kwargs)


def array(description, **kwargs) -> typing.Type:
    """Create a :class:`~doctor.types.Array` type.

    :param description: A description of the type.
    :param kwargs: Can include any attribute defined in
        :class:`~doctor.types.Array`
    """
    kwargs['description'] = description
    return type('Array', (Array,), kwargs)


def new_type(cls, **kwargs) -> typing.Type:
    """Create a user defined type.

    The new type will contain all attributes of the `cls` type passed in.
    Any attribute's value can be overwritten using kwargs.

    :param kwargs: Can include any attribute defined in
        the provided user defined type.
    """
    props = dict(cls.__dict__)
    props.update(kwargs)
    return type(cls.__name__, (cls,), props)



from functools import wraps

import mock
from doctor.flask import create_routes
from doctor.routing import get, Route

from .base import FlaskTestCase
from .types import ItemId, Name


def adds_positional_arg_to_func(arg=1):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            return func(arg, *args, **kwargs)
        return wrapper
    return decorator


@adds_positional_arg_to_func(55)
def logic_func(pos_arg, item_id: ItemId, name: Name=None):
    return {
        'item_id': item_id,
        'name': name,
        'pos_arg': pos_arg,
    }


class RouterIntegrationTestCase(FlaskTestCase):

    before = mock.Mock()
    after = mock.Mock()

    def get_routes(self):
        routes = (
            Route('/test/', methods=[
                get(logic_func)], heading='Test', before=self.before,
                after=self.after),
        )
        return create_routes(routes)

    def test_decorated_logic_func_passes_arg_to_func(self):
        """
        This test verifies everything works properly if our logic function
        is decorated and the decorator passes an extra positional argument to
        it.  This was broken in v1.1.3 as it would cause a 400 saying that
        `pos_arg` is required as a request parameter since it was not taking
        into accout `omit_args` and just determining required arguments based
        on the function signature's non-keyword arguments.
        """
        response = self.client.get('/test/', query_string={'item_id': 1})
        expected = {
            'item_id': 1,
            'name': None,
            'pos_arg': 55,
        }
        assert expected == response.json
        assert self.before.called
        assert self.after.called
